#include <iostream>
#include "assert.h"

#include "compile.hh"
#include "symtables.hh"
#include "binding.hh"
#include "tristate"

#define CPP_LBOUND_PARAM_NAME(x) (std::string(x)+"_ML")
#define CPP_DIMENSION_PARAM_NAME(x) (std::string(x)+"_MS")
#define CPP_SUBFUNC_NAME(x) (std::string(x)+"_f")
#define CPP_MEMBER_NAME(x) (std::string(x))
#define CPP_UTYPE_NAME(x)  (std::string(x)+"_u")
#define CPP_INT_NAME(x)    (std::string(x)+"_i")
#define CPP_SINGLE_NAME(x) (std::string(x)+"_z")
#define CPP_DOUBLE_NAME(x) (std::string(x)+"_d")
#define CPP_LONG_NAME(x)   (std::string(x)+"_l")
#define CPP_STRING_NAME(x) (std::string(x)+"_s")
#define CPP_ARRAY_NAME(x)  (std::string(x)+"_a")

std::vector<std::string> global_data;

static const std::pair<std::string, std::string>
    GetTypeBaseNames(const TypeInfo& type, const std::string& name)
{
    switch(type.Type)
    {
        case TypeInfo::Integer: return std::make_pair("BAS_INT", CPP_INT_NAME(name));
        case TypeInfo::Single:  return std::make_pair("BAS_SGN", CPP_SINGLE_NAME(name));
        case TypeInfo::Double:  return std::make_pair("BAS_DBL", CPP_DOUBLE_NAME(name));
        case TypeInfo::Long:    return std::make_pair("BAS_LNG", CPP_LONG_NAME(name));
        case TypeInfo::String:
            return type.fixlength
                ? /* FIXME */
                  std::make_pair("BAS_STR", CPP_STRING_NAME(name))
                : std::make_pair("BAS_STR", CPP_STRING_NAME(name));
        case TypeInfo::UserType:
            return std::make_pair(type.utypename, CPP_UTYPE_NAME(name));
    }
    /* not reached, but keeps compiler happy */
    assert(!"not_reached");
    return std::make_pair(type.utypename, CPP_UTYPE_NAME(name));
}
static const std::string CompileStringConst(const std::string&str)
{
    std::string result;
    result += '"';
    for(unsigned a=0; a<str.size(); ++a)
    {
        unsigned char c = str[a];
        if(c=='\n') { result += "\\n"; continue; }
        if(c=='\r') { result += "\\r"; continue; }
        if(c=='\t') { result += "\\t"; continue; }
        if(c=='\b') { result += "\\b"; continue; }
        if(c=='\a') { result += "\\a"; continue; }
        if(c=='"' || c=='\\') { result += '\\'; result += str[a]; continue; }
        if(c < 32) { char Buf[64]; sprintf(Buf, "\\%03o", c); result += Buf; continue; }
        result += str[a];
    }
    result += '"';
    return result;
}

enum VarDeclType { Init_Var, Param_Ref, Param_Copy };

static struct
{
    void operator() (std::ostream& out, const std::string& name,
                     TypeP typep, VarDeclType decl_type = Init_Var
                    )
    {
        assert(typep);
        const TypeInfo& type = *typep;

        const std::pair<std::string, std::string> base = GetTypeBaseNames(type, name);

        switch(type.arraytype)
        {
            case TypeInfo::NotArray:
            {
                out << base.first << ' ';
                if(decl_type == Param_Ref) out << "&";
                out << base.second;
                break;
            }
            case TypeInfo::StaticArray:
            {
                out << base.first << ' ';
                if(decl_type == Param_Copy)
                {
                    fprintf(stderr, "Warning: BYVAL Array parameters not supported\n");
                }
                out << CPP_ARRAY_NAME(base.second);
                for(unsigned a=0; a<type.dimensions.size(); ++a)
                    out << (!a ? "[" : "][") << type.dimensions[a].second;
                out << "]";
                break;
            }
            case TypeInfo::DynamicArray:
            {
                out << "BAS_DYNARRAY<" << base.first << ", " << type.dimensions.size() << "> ";
                if(decl_type == Param_Ref) out << "&";
                out << CPP_ARRAY_NAME(base.second);
                
                if(decl_type == Init_Var)
                {
                    for(unsigned a=0; a<type.dimensions.size(); ++a)
                    {
                        out << (!a ? "(" : ", ")
                            << type.dimensions[a].first
                            << ","
                            << type.dimensions[a].second;
                    }
                    out << ")";
                }
                break;
            }
            case TypeInfo::ParamArray:
            {
                out << "BAS_MAGICARRAY<" << base.first << ">";
                if(decl_type == Param_Ref) out << "&";
                out << CPP_ARRAY_NAME(base.second);
                break;
            }
        }
    }
private:
} CompileVarDeclaration;

static void CompileTypeDefs(const Scope& scope)
{
    for(hash_map<std::string, UtypeP>::const_iterator
        i = scope.types.begin();
        i != scope.types.end();
        ++i)
    {
        UtypeP p = i->second;
        assert(p);
        
        std::cout << "struct " << i->first << " {\n";
        
        for(unsigned a=0; a<p->size(); ++a)
        {
            std::cout << "  ";
            CompileVarDeclaration(std::cout, CPP_MEMBER_NAME((*p)[a].first), (*p)[a].second );
            std::cout << ";\n";
        }
        
        std::cout << "};\n";
    }
}

static void CompileVars(const ScopeVarListType& vars, tristate shared_only, bool implicit_static=false)
{
    for(ScopeVarListType::const_iterator i = vars.begin(); i != vars.end(); ++i)
    {
        ctxp ctx = i->second;
        assert(ctx);
        if(!shared_only.is_maybe() && ctx->is_shared != shared_only) continue;

        if(ctx->is_static || implicit_static) std::cout << "static ";
        
        CompileVarDeclaration(std::cout, i->first, ctx->type);
        std::cout << ";\n";
    }
}

static void CompileVars(const Scope& scope, tristate shared_only, bool implicit_static=false)
{
    CompileVars(scope.Utype  .scalars, shared_only, implicit_static);
    CompileVars(scope.Utype  .arrays, shared_only, implicit_static);
    CompileVars(scope.Int    .scalars, shared_only, implicit_static);
    CompileVars(scope.Int    .arrays, shared_only, implicit_static);
    CompileVars(scope.Single .scalars, shared_only, implicit_static);
    CompileVars(scope.Single .arrays, shared_only, implicit_static);
    CompileVars(scope.Double .scalars, shared_only, implicit_static);
    CompileVars(scope.Double .arrays, shared_only, implicit_static);
    CompileVars(scope.Long   .scalars, shared_only, implicit_static);
    CompileVars(scope.Long   .arrays, shared_only, implicit_static);
    CompileVars(scope.String .scalars, shared_only, implicit_static);
    CompileVars(scope.String .arrays, shared_only, implicit_static);
}

static void CompileSubFunc(const std::string& name, const SubRoutine& sub, bool decl_only)
{
    std::cout << "static ";
    if(sub.returntype)
    {
        const std::pair<std::string, std::string> base
            = GetTypeBaseNames(*sub.returntype, name);
        std::cout << "const " << base.first;
    }
    else
        std::cout << "void";
    std::cout << " " << CPP_SUBFUNC_NAME(name) << "(";
    
    for(unsigned a=0; a<sub.params.size(); ++a)
    {
        const SubRoutine::ParamSpec& param = sub.params[a];
        if(a) std::cout << ", ";
        CompileVarDeclaration(std::cout,
                              param.name,
                              param.type,
                              param.passconvention ==
                                SubRoutine::ParamSpec::ByVal
                                    ? Param_Copy
                                    : Param_Ref);
        if(param.is_array)
        {
            std::cout << ", BAS_INT " << CPP_LBOUND_PARAM_NAME(param.name);
            std::cout << ", BAS_INT " << CPP_DIMENSION_PARAM_NAME(param.name);
        }
    }
    
    std::cout << ")";
    
    if(decl_only)
        std::cout << ";\n";
    else
    {
        std::cout << "\n";
    }
}

static void CompileSubDefs(const Scope& scope)
{
    for(hash_map<std::string, SubFuncP>::const_iterator
        i = scope.subroutines.begin(); i != scope.subroutines.end(); ++i)
    {
        SubFuncP sub = i->second; assert(sub);
        CompileSubFunc(i->first, *sub, true);
    }
}

static void CompileCode(stmtp codep, const Scope& scope)
{
    if(!codep) return;
    CompileCode(codep->get_prev(), scope);
    codep->Compile(scope);
}

unsigned TranslateLabelToDataPtr(const long double l)
{
    // TODO: Use current scope and parent scope
    return 0;
}

static void VerifyLabels(const Scope& scope)
{
    for(std::map<long double, LabelDef>::const_iterator
        i = scope.labels.begin(); i != scope.labels.end(); ++i)
    {
        if(!i->second.defined)
            fprintf(stderr, "Error: Label %lg not defined\n", i->first);
    }
}

/*
PLAN:
    Define typedefs. - DONE
    Define global variables. -SHARED ONLY DONE
    Declare subroutines. -DONE
    
    Compile code.
    
    Compile each function.

NOTES:
    Constants are not defined. They are compiled!

TODO:
    SHARED imports?
    
*/
void Compile(stmtp main_codep,
             const std::vector<std::string>& global_data)
{
    ScopeP scope = GetRootScope();
    VerifyLabels(*scope);
    
    std::cout << "static const unsigned DATA_COUNT = " << global_data.size() << ";\n";
    if(!global_data.empty())
    {
        std::cout << "static unsigned DATA_PTR = 0;\n";
        std::cout << "static const char* const DATA_TABLE[DATA_COUNT] = {\n";
        for(unsigned a=0; a<global_data.size(); ++a)
        {
            //std::cout << "// " << a << "\n";
            std::cout << CompileStringConst(global_data[a]);
            if(a+1 < global_data.size()) std::cout << ",";
            std::cout << "\n";
        }
        std::cout << "}\n";
    }
    
    CompileTypeDefs(*scope);
    CompileVars(*scope, true, true); // shared only
    CompileSubDefs(*scope);

    std::cout << "int main(int argc, char**argv)\n";
    std::cout << "{\n";
    CompileVars(*scope, false, false); // local only
    CompileCode(main_codep, *scope);
    std::cout << "}\n";

    /* Compile subs */
    for(hash_map<std::string, SubFuncP>::const_iterator
        i = scope->subroutines.begin(); i != scope->subroutines.end(); ++i)
    {
        SubFuncP sub = i->second; assert(sub);
        ScopeP subscope = sub->sub_scope;
        VerifyLabels(*subscope);
        
        CompileSubFunc(i->first, *sub, false);
        std::cout << "{\n";
        CompileVars(*subscope, false, sub->is_static); // local only
        CompileCode(sub->code, *subscope);
        std::cout << "}\n";
    }
}

void stmt_if::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_select::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_doloop::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_for::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_return::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_goto::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_gosub::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_restore::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_read::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_let::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_print::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_input::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_fileopen::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_redim::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_subcall::Compile(const Scope& scope) { stmt::Compile(scope); }
void stmt_label::Compile(const Scope& scope) { stmt::Compile(scope); }
